import datetime
import numpy as np
import sys


from constants import *
from electrochem import electrochem
from fluid import fluid
import result_io
from solver import solver
from space import space
from space import boundary
from space import source
from thermo import thermo

if DEBUG:
    from debug import *


#######################
### create obejects ###

simSpace = space()
solver_obj = solver()
thermoObj = thermo()
electrochemObj = electrochem()
boundaries = dict()
#tempspace
boundaries['insulated'] = boundary('N',0)
boundaries['T_gas_in_a'] = boundary('D',T_GAS_IN_A)
boundaries['T_gas_in_c'] = boundary('D',T_GAS_IN_C)

boundaries['m_H2O_in_a'] = boundary('D',MASS_H2O_IN_ANODE)
boundaries['m_H2O_in_c'] = boundary('D',MASS_H2O_IN_CATHODE)
boundaries['m_H2_in'] = boundary('D',MASS_H2_IN_CATHODE)
boundaries['m_O2_in'] = boundary('D',MASS_O2_IN_ANODE)

thermal_sources:list[source] = list()
#thermal_sources.append(source(round(ROWPTS/2)-1,round(COLPTS/2)-1,10/(dy[0,0]*dx[0,0]))) # in W/m^2 for 2D
simSpace.tempspace.setSources(thermal_sources)


#######################
### run simulation  ###


# Print general simulation information
print("######## Beginning Simulation ########")
print("######################################")
print("# Timestemp: {0}".format(datetime.datetime.now()))
print("# Max Iterations: {0}".format(MAX_ITERATIONS))
print("# Mesh: {0} x {1}".format(COLPTS,ROWPTS))
#print("# Re/u: {0:.2f}\tRe/v:{1:.2f}".format(RHO*LENGTH_X/MU,RHO*LENGTH_Y/MU))
if PRINT_TO_FILE:
    print("# Save outputs to text file: {0}".format(bool(RESULTS_DIRECTORY)))

# Make directory to store results
result_io.makeDirectory()

# Initialize counters
time = float(0)
iteration = 0

epsilon_t = float(300.0)

if DEBUG:
    mean_iteration_time = float(0)

# Run simulation
while(iteration < MAX_ITERATIONS):
    if DEBUG:
        start_iteration_time = datetime.datetime.now()
    # Solve timestep
    epsilon_t, epsilon_i, dt = solver_obj.solveTimestep(simSpace,thermoObj,electrochemObj,boundaries,epsilon_t)
    
    #Print time left
    outstr = f"\rIteration: {iteration}; T_min: {simSpace.tempspace.t_c.min():.2f}, T_max: {simSpace.tempspace.t_c.max():.2f}; i_mean: {np.mean(simSpace.current_desity[simSpace.current_desity!=0])/1e4:.3f}; dt: {dt:.4E}; epsilon_T: {epsilon_t:.4E}, epsilon_i: {epsilon_i:.4E}"
    if not DEBUG:
        sys.stdout.write(outstr)
        sys.stdout.flush()
    else:
        print(outstr)
        print("\n##################################################\n")
    
    if PRINT_TO_FILE:
        result_io.writeToFile(simSpace,iteration)

    if DEBUG:
        #print memory usage
        for name, size in sorted(((name, sys.getsizeof(value)) for name, value in list(
                                locals().items())), key= lambda x: -x[1])[:10]:
            print("{:>30}: {:>8}".format(name, sizeof_fmt(size)))
        iteration_time = (datetime.datetime.now() - start_iteration_time).total_seconds() * 10**3
        mean_iteration_time = (mean_iteration_time * iteration + iteration_time) / (iteration + 1)
        print(f"Execution time of iteration: {iteration_time} ms; Mean iteration time: {mean_iteration_time} ms")

    #test if converged
    if epsilon_t < TOLERANCE_TEMPERATURE and epsilon_i < TOLERANCE_CURRDENS :
        print("\n######################################")
        print("Simulation converged!")
        print("# Timestemp: {0}".format(datetime.datetime.now()))
        print(f"Time: {time}")
        print(f"Iterations: {iteration}")
        print(f"Mean Current Density: {np.mean(simSpace.current_desity[simSpace.current_desity!=0])/1e4} A/cm^2")
        print(f"Mean Cell Voltage: {np.mean(simSpace.voltage)} V")

        #calculate energy imbalace: 
        sources = simSpace.tempspace.s[1:ROWPTS+1, 1:COLPTS+1] + thermoObj.calcHeatSourcesFluidsFullyEquilibriated(simSpace) + thermoObj.calcHeatFlowToAmbient(simSpace) + electrochemObj.heat_overpotential
       
        if NUMBER_CELLS >1:
            sources_sample_cell = sources[np.array(INDEX_ANODE_FLOWFIELD)[int(NUMBER_CELLS/2)]-1:np.array(INDEX_ANODE_FLOWFIELD)[int(NUMBER_CELLS/2)+1]-1,:] # W/cell
        else:
            sources_sample_cell = sources[np.array(INDEX_ANODE_FLOWFIELD)[0]-1:np.array(INDEX_ANODE_FLOWFIELD)[0]+N_CELL_LAYERS-1,:] # W/cell
            
        #sources_sample_cell *= simSpace.area[np.array(INDEX_ANODE_FLOWFIELD)[int(NUMBER_CELLS/2)]:np.array(INDEX_ANODE_FLOWFIELD)[int(NUMBER_CELLS/2)+1],1:COLPTS+1] # W abolute
        print(f"Energy imbalace (one EC): {np.sum(sources_sample_cell)}W (max.source: {np.max(sources_sample_cell)}W/mesh-cell)")

        # Calculate efficiency (Voltage efficiency according to higher heating value)
        u_thn = electrochemObj.calcThermoneutralCellPotential(simSpace)
        efficiency = np.mean(u_thn) / np.mean(simSpace.voltage)
        print(f"Voltage efficiency: {efficiency}")
                   
        break

    time += dt
    iteration += 1 
result_io.writeToFile(simSpace,iteration,isFinal=True)
print("\n######################################")
print(f"Epsilon_t: {epsilon_t}")
print(f"t_min: {simSpace.tempspace.t_c.min()}, t_max: {simSpace.tempspace.t_c.max()}")