import numpy as np

from electrochem import electrochem
from space import space
from constants import *
from thermo import thermo


class solver:

    def __init__(self) -> None:
        self.multiplier = float(1) # for solve timestep calculation 
        pass

    #def solveTimestep(self, space:space, fluid:fluid, boundaries:dict) -> float:
    def solveTimestep(self, space:space, thermo:thermo, electrochem: electrochem, boundaries:dict, epsilon_t:float) -> tuple[float,float,float]:
        # solves one timestep of the simulaton 
        # Args:
        #   space: the space object 
        #   thermo: object of thermo class
        #   electrochem: object of electrochem class 
        #   boundary: dict of boundary condition objet for the variables 
        #   epsilon_t: the maximum temperature difference in the last timestep
        # Return: 
        #   epsilon_t: maximum temperature difference to last timestep
        #   timestep: the size of the calculated timestep
        
        # calc timestep
        self.setTimeStep(CFL_NUMBER,space, epsilon_t)

        ### set boundary values ###

        # # flowspace
        # space.flowspace.setUBoundary(boundaries['noslip'], boundaries['flow'], boundaries['noslip'], boundaries['noslip'])
        # space.flowspace.setVBoundary(boundaries['noslip'], boundaries['noslip'], boundaries['noslip'], boundaries['noslip'])
        # space.flowspace.setPBoundary(boundaries['zeroflux'], boundaries['pressureatm'], boundaries['zeroflux'],boundaries['zeroflux'])

        # temperature space
        space.tempspace.setTBoundary(boundaries['insulated'],boundaries['insulated'],boundaries['insulated'],boundaries['insulated']) #Bock.2020 sais adiabatic
        
        space.tempspace.t[INDEX_ANODE_FLOWFIELD,0] = boundaries['T_gas_in_a'].value
        if PARALLEL_FLOW:
            space.tempspace.t[INDEX_CATHODE_FLOWFIELD_BACKSIDE,0] = boundaries['T_gas_in_c'].value
        else:
            space.tempspace.t[INDEX_CATHODE_FLOWFIELD_BACKSIDE,-1] = boundaries['T_gas_in_c'].value
        space.flowspace.m_y[INDEX_ANODE_FLOWFIELD,0,0] = boundaries['m_H2O_in_a'].value
        space.flowspace.m_y[INDEX_ANODE_FLOWFIELD,0,2] = boundaries['m_O2_in'].value
        if PARALLEL_FLOW:    
            space.flowspace.m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,0,1] = boundaries['m_H2_in'].value
            space.flowspace.m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,0,0] = boundaries['m_H2O_in_c'].value
        else:
            space.flowspace.m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,-1,1] = -boundaries['m_H2_in'].value
            space.flowspace.m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,-1,0] = -boundaries['m_H2O_in_c'].value



        ### solve step ###

        # # flowspace
        # # calc starred velocities
        # self.getStarredVelocities(space,fluid)
        # # solve pressure poisson
        # self.solvePressurePoisson(space, fluid)
        # space.flowspace.setPBoundary(boundaries['zeroflux'], boundaries['zeroflux'], boundaries['pressureatm'],boundaries['zeroflux'])
        # #solve momentum equation
        # self.solveMomentumEquation(space,fluid)


        i = electrochem.calcCurrentDensity(space, U_STACK / NUMBER_CELLS, space.current_desity, thermo)

        epsilon_i = np.abs((i - space.current_desity)).max()
        space.current_desity = i

        
        self.solveMassflowEquation(space,thermo)
        # temperature space
        self.solveTemperatureEquation(space,thermo, electrochem)


        # Set center variables
        space.flowspace.setCenter()
        space.tempspace.setCenter()

        # check if converged
        epsilon_t = np.abs(space.tempspace.t_c - space.tempspace.t_c_last).max()

        return epsilon_t, epsilon_i, self.dt
    


    def setTimeStep(self, CFL, space:space, epsilon_t:float) -> None:
        # Determines the timesteps for the simulation
        # Args:
        #   CFL: Courant–Friedrichs–Lewy criterion as upper bound for timesteps -> Ensures that information propagated in one timestep is not farther than distance between two meshes
        #   space: the space object 
        #   epsilon_t: the maximum temperature difference within the last timestep

        
        # make shure that stablity critera is met (https://en.wikipedia.org/wiki/FTCS_scheme)

        # if epsilon_t < THRESHOLD_TIMESTEP_INCREASE:
        #     self.multiplier = 1.1*self.multiplier
        # if epsilon_t > THRESHOLD_TIMESTEP_DECREASE:
        #     self.multiplier = 0.9*self.multiplier
        #factor = 4* np.max([np.max(space.tempspace.k_x/(space.dx**2)), np.max(space.tempspace.k_y/(space.dy**2))])
        # factor = 4* (1/np.min(space.dx)**2+1/np.min(space.dy)**2)
        factor = 4* (np.max(space.tempspace.k_x/(space.dx**2)) + np.max(space.tempspace.k_y/(space.dy**2)))
        
        dt = CFL / factor * self.multiplier

        self.dt = dt


    #####################################################
    ################ Solver funcitons ###################

    # flowspace
    def solveMassflowEquation(self, space:space, thermo:thermo) -> None:
        # calculates the massflow at timestep t+delta_t 
        # Args:
        #   space: space object
        #   thermo: thermo object

        m_x = space.flowspace.m_x.astype(float,copy=False)
        m_y = space.flowspace.m_y.astype(float,copy=False)
        
        h2o_used = space.current_desity * CELLAREA_SEGMENT / (2 * FARRADAY) * MOLAR_MASS_H2O
        n_dragg = 2.5 * LAMBDA / 22 # Springer.1991
        if INCLUDE_DRAG:
            h2o_crossover = space.current_desity * CELLAREA_SEGMENT / (FARRADAY) * MOLAR_MASS_H2O * n_dragg # Approach of Springer.1991 (no division by 2 here, as reffered to hydrogen ION)
        else:
            h2o_crossover = 0 
        h2_produced = space.current_desity * CELLAREA_SEGMENT / (2*FARRADAY) * MOLAR_MASS_H2
        o2_produced = space.current_desity * CELLAREA_SEGMENT / (4*FARRADAY) * MOLAR_MASS_O2

        # Anode flowfield
        m_y[INDEX_ANODE_FLOWFIELD,1:COLPTS+1,0] = m_y[INDEX_ANODE_FLOWFIELD, 0:COLPTS,0] 
        m_y[INDEX_ANODE_FLOWFIELD,2:,0] -= (h2o_used + h2o_crossover)
        m_y[INDEX_ANODE_FLOWFIELD,1:COLPTS+1,2] = m_y[INDEX_ANODE_FLOWFIELD, 0:COLPTS,2] 
        m_y[INDEX_ANODE_FLOWFIELD,2:,2] += o2_produced
        m_y[INDEX_ANODE_FLOWFIELD,-1,:] = m_y[INDEX_ANODE_FLOWFIELD, -2,:]
        if np.any(m_y[INDEX_ANODE_FLOWFIELD,:,0]<0):
            raise Exception("Not sufficient water in anode!")
        # Cathode flowfield
        if PARALLEL_FLOW:
            m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,1:COLPTS+1,1] = m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE, 0:COLPTS,1]
            m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,2:,1] += h2_produced
            m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,1:COLPTS+1,0] = m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE, 0:COLPTS,0] 
            m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,2:,0] += h2o_crossover
            m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,-1,:] = m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE, -2,:]

        else:
            m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,0:COLPTS+1,1] = m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE, 1:COLPTS+2,1]
            m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,1:COLPTS+1,1] -= h2_produced
            m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,0:COLPTS+1,0] = m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE, 1:COLPTS+2,0] 
            m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,1:COLPTS+1,0] -= h2o_crossover # minus because flow in negative y direction
        # flow in x has to be here but not in anode flowfield as point at first cellborder has cellindex but point at last cellborder has next index
        m_x[INDEX_CATHODE_FLOWFIELD,1:COLPTS+1,0] = h2o_crossover            
        m_x[INDEX_CATHODE_FLOWFIELD,1:COLPTS+1,1] = h2_produced
        # m_x[INDEX_CATHODE_FLOWFIELD_BACKSIDE,1:COLPTS+1,0] = h2o_crossover            
        # m_x[INDEX_CATHODE_FLOWFIELD_BACKSIDE,1:COLPTS+1,1] = h2_produced


        # GDL
        if HAS_GDL:
            m_x[INDEX_GDL_A,1:COLPTS+1,0] = h2o_used + h2o_crossover
            m_x[INDEX_GDL_A,1:COLPTS+1,2] = -o2_produced             
            m_x[INDEX_GDL_C,1:COLPTS+1,0] = h2o_crossover            
            m_x[INDEX_GDL_C,1:COLPTS+1,1] = h2_produced

        # CL
        m_x[INDEX_ACL,1:COLPTS+1,0] = h2o_used + h2o_crossover
        m_x[INDEX_ACL,1:COLPTS+1,2] = -o2_produced 
        m_x[INDEX_CCL,1:COLPTS+1,0] = h2o_crossover
        m_x[INDEX_CCL,1:COLPTS+1,1] = h2_produced

        # Inner FF boundary Layers
        m_x[INDEX_INNER_ANODE_FF_BOUNDARY_LAYER,1:COLPTS+1,0] = h2o_used + h2o_crossover            
        m_x[INDEX_INNER_ANODE_FF_BOUNDARY_LAYER,1:COLPTS+1,2] = -o2_produced 
        m_x[INDEX_INNER_CATHODE_FF_BOUNDARY_LAYER,1:COLPTS+1,0] = h2o_crossover            
        m_x[INDEX_INNER_CATHODE_FF_BOUNDARY_LAYER,1:COLPTS+1,1] = h2_produced
       

        # Membrane
        m_x[INDEX_MEMBRANE,1:COLPTS+1,0] = h2o_crossover
        m_x[INDEX_MEMBRANE,1:COLPTS+1,1] = h2_produced

        # calculate amount of gasous water
        space.flowspace.m_x_h2o_g = thermo.calcMassFlowGasousWaterInH2O2(space,m_x[:,:,1], m_x[:,:,2])
        space.flowspace.m_y_h2o_g = thermo.calcMassFlowGasousWaterInH2O2(space,m_y[:,:,1], m_y[:,:,2])

        # check mass balance
        #if DEBUG:
            # print(f'H2O_unbalance:{m_x[2:,1:COLPTS+1,0]+m_y[1:ROWPTS+1,2:,0]-m_x[1:ROWPTS+1,1:COLPTS+1,0]-m_y[1:ROWPTS+1,1:COLPTS+1,0]}')
            # print(f'H2_unbalance:{m_x[2:,1:COLPTS+1,1]+m_y[1:ROWPTS+1,2:,1]-m_x[1:ROWPTS+1,1:COLPTS+1,1]-m_y[1:ROWPTS+1,1:COLPTS+1,1]}')
            # print(f'O2_unbalance:{m_x[2:,1:COLPTS+1,2]+m_y[1:ROWPTS+1,2:,2]-m_x[1:ROWPTS+1,1:COLPTS+1,2]-m_y[1:ROWPTS+1,1:COLPTS+1,2]}')

         

    # temperaturespace
    def solveTemperatureEquation(self, space:space, thermo:thermo, electrochem:electrochem) -> None:
        # calculate the temperatures at timestep t+delta_t with the forward Euler method
        # Args:
        #   space: space object
        #   thermo: thermo object
        #   electrochem: electrochem object 

        rows = int(space.rowpts)
        cols = int(space.colpts)
        k_x = space.tempspace.k_x.astype(float,copy=False)
        k_y = space.tempspace.k_y.astype(float,copy=False)
        temp   = space.tempspace.t.astype(float,copy=False)
        dx = space.dx.astype(float,copy=False)
        dy = space.dy.astype(float,copy=False)

        # use forward Euler method
    
        # calculate laplacian
        laplacian_x = ((k_x[1:rows+1,1:cols+1]*(temp[2:,1:cols+1]-temp[1:rows+1,1:cols+1])/dx[1:rows+1,1:cols+1])-(k_x[0:rows,1:cols+1]*(temp[1:rows+1,1:cols+1]-temp[0:rows,1:cols+1])/dx[0:rows,1:cols+1]))/(dx[1:rows+1,1:cols+1])
        
        if not IS_1D:
            laplacian_y = ((k_y[1:rows+1,1:cols+1]*(temp[1:rows+1,2:]-temp[1:rows+1,1:cols+1])/dy[1:rows+1,1:cols+1])-(k_y[1:rows+1,0:cols]*(temp[1:rows+1,1:cols+1]-temp[1:rows+1,0:cols])/dy[1:rows+1,0:cols]))/(dy[1:rows+1,1:cols+1])
        else:
            laplacian_y = 0
        #only for static state; For dynamic simulation devide laplacian and source term by density*cp
        #temp[1:rows+1,1:cols+1] = temp[1:rows+1,1:cols+1] + (laplacian_x + laplacian_y) * self.dt + (thermo.calcHeatSourcesFluidsFullyEquilibriated(space)) * self.dt
        sources = space.tempspace.s[1:rows+1, 1:cols+1] + thermo.calcHeatSourcesFluidsFullyEquilibriated(space) + thermo.calcHeatFlowToAmbient(space) + electrochem.heat_overpotential


        temp[1:rows+1,1:cols+1] = temp[1:rows+1,1:cols+1] + (laplacian_x + laplacian_y) * self.dt + (sources / space.area[1:ROWPTS+1,1:COLPTS+1] / CELLSIZE_Z) * self.dt

        space.heat_flux[:,:,0] = -(temp[1:rows+1,1:cols+1] - temp[0:rows,1:cols+1]) * k_x[0:rows, 1:cols+1] / space.dx[0:rows,1:cols+1]  #in W/m^2(cellarea)
        space.heat_flux[:,:,1] = -(temp[1:rows+1,1:cols+1] - temp[1:rows+1,0:cols]) * k_y[1:rows+1, 0:cols] / space.dy[1:rows+1,0:cols] #in W/m^2(cellarea)
        # apply positive heat fluxes to left side of cell (and negative to right side of cell)
        __heat_flux_pos = space.heat_flux.copy()
        __heat_flux_pos[__heat_flux_pos<=0]=0
        space.heat_flux -= __heat_flux_pos
        space.heat_flux[:-1,:,0] += __heat_flux_pos[1:,:,0]
        space.heat_flux[:,:-1,1] += __heat_flux_pos[:,1:,1]
        
        
        if DEBUG:
            #print(f"Temperature: {temp-273.15}°C")
            #calc mass imbalance for cell in middle of stack 
            if NUMBER_CELLS >1:
                print(f"Energy imbalace (one cell): {np.sum(sources[np.array(INDEX_ANODE_FLOWFIELD)[int(NUMBER_CELLS/2)]-1:np.array(INDEX_ANODE_FLOWFIELD)[int(NUMBER_CELLS/2)+1]-1,:])}W")
            else:             
                print(f"Energy imbalace (one cell): {np.sum(sources[np.array(INDEX_ANODE_FLOWFIELD)[0]-1:np.array(INDEX_ANODE_FLOWFIELD)[0]+N_CELL_LAYERS-1,:])}W")
                
     