import numpy as np
from constants import *


class boundary:
#contains boundary condition Dirichlet (defined value) or Neumann (defined derivative)
    def __init__(self, boundary_type:str , boundary_value:float) -> None:
        self.defineBoundary(boundary_type,boundary_value)

    def defineBoundary(self, boundary_type:str, boundary_value:float) -> None:
        # Sets the boundary conditions
        # Args:
        #   boundary_type: "N" for Neumann or "D" for Dirichlet
        #   boundary_value: value of the boundary condition (W/m^2 for heat sources in 2D)

        self.type = boundary_type
        self.value = boundary_value

class source:
#contains sources
    def __init__(self, x_pos:int, y_pos:int, value:float) -> None:
        # sets the parameters of the source
        # Args: 
        #   x_pos: position of the source in x direction
        #   y_pos: position of the source in y direction
        #   value: strength of the source
        
        self.x = x_pos
        self.y = y_pos
        self.value = value

class space: 
    # Handles the mesh 

    def __init__(self) -> None:
        #initlialize spaces of the different physics
        self.tempspace = temperatureSpace(self)
        self.flowspace = flowSpace(self)

        self.createMesh()

        self.tempspace.initTemperature(INITIAL_TEMPERATURE)
        self.flowspace.initMassflow()
        self.current_desity = np.ones_like(self.tempspace.t[INDEX_MEMBRANE,1:COLPTS+1]) * INITAL_CURRENT_DENSITY
        self.current_desity[:,-1] = 0 #outlet of last cell 
        self.voltage = np.zeros_like(self.current_desity)
        self.heat_flux = np.zeros((ROWPTS,COLPTS,2))

    def createMesh(self, rowpts:int=ROWPTS, colpts:int=COLPTS):
        # Creates the mesh in 2D with x-columns and y-rows
        # Args:
        #   rowpts: number of cells in row direction
        #   colpts: number of cells in col direction

        self.rowpts = rowpts
        self.colpts = colpts

        self.setDeltas()

        # create meshes of physics
        #self.flowspace.createMesh(rowpts, colpts)
        self.tempspace.createMesh(rowpts,colpts)
        self.flowspace.createMesh(rowpts,colpts)


    def setDeltas(self) -> None:
        # Sets the size of the cells 

        dx = np.ones((ROWPTS+2,COLPTS+2))
        dy = np.ones((ROWPTS+2,COLPTS+2))
        dx[1:ROWPTS+1, :] = np.full((COLPTS+2,ROWPTS),CELLSIZE_X).transpose()
        dx[0,:] = CELLSIZE_X[-1] # set to size of virtual cells
        dx[-1,:] = CELLSIZE_X[-1] # cellsize not used but must be bigger than minimal cell size not to influence dt
        dy[:, 1:COLPTS+1] = np.full((ROWPTS+2,COLPTS),CELLSIZE_Y)
        dy[:, 0] = CELLSIZE_Y[-1] # set to size of virtual cells
        dy[:, -1] = CELLSIZE_Y[-1] # cellsize not used but must be bigger than minimal cell size not to influence dt



        self.area = dx*dy
        self.dx=dx
        self.dy=dy



class flowSpace: 
    def __init__(self, space:space) -> None:
        self.space = space

    # hanldes the variables in the mesh regarding the flow
    def createMesh(self, rowpts, colpts:int) -> None: 
        # Creates the mesh
        # Args:
        #   rowpts: number of cells in row direction
        #   colpts: number of cells in col direction
        self.rowpts = rowpts
        self.colpts = colpts

        # create mass-flow matrix (with 3rd dimension = type of gas (Water(total), H2, O2))
        self.m_x = np.ones((self.rowpts+2, self.colpts+2,3)) # in kg/s 
        self.m_y = np.ones((self.rowpts+2, self.colpts+2,3)) # in kg/s
        self.m_x_c = np.ones((self.rowpts, self.colpts,3)) # in kg/s 
        self.m_y_c = np.ones((self.rowpts, self.colpts,3)) # in kg/s
        self.m_x_h2o_g =  np.ones((self.rowpts+2, self.colpts+2)) # in kg/s; mass-flow h2o gas (the maximum gasous massflow at given conditions)
        self.m_y_h2o_g =  np.ones((self.rowpts+2, self.colpts+2)) # in kg/s; mass-flow h2o gas (the maximum gasous massflow at given conditions)
        self.m_x_h2o_c = np.ones((self.rowpts, self.colpts,2)) # in kg/s; 0: liquid, 1: gas (used to store the actual massflows of gas)
        self.m_y_h2o_c = np.ones((self.rowpts, self.colpts,2)) # in kg/s; 0: liquid, 1: gas (used to store the actual massflows of gas)


        # create enthalpy-flow matrix 
        # self.h_x = np.zeros((self.rowpts+2, self.colpts+2)) # in J/kg
        # self.h_y = np.zeros((self.rowpts+2, self.colpts+2)) # in J/kg

        # initialize source term
        # self.setSourceTermEnthalpie()
        self.setSourceTermMassflow()

    # def init_enthalpy(self, H:float) -> None:
    #     # Initializes the specific enthalpy in all cells 
    #     # Args:
    #     #   H: Initial value to be set for the specific enthalpy

    #     self.h_x = H*self.h_x
    #     self.h_y = H*self.h_y

    def initMassflow(self) -> None:
        # Initializes the massflow in all cells (must be zero)

        self.m_x = 0*self.m_x
        self.m_y = 0*self.m_y 
        self.m_x_h2o_g = 0*self.m_x_h2o_g
        self.m_y_h2o_g = 0*self.m_y_h2o_g
        self.m_y[INDEX_ANODE_FLOWFIELD,:,0] = MASS_H2O_IN_ANODE
        self.m_y[INDEX_ANODE_FLOWFIELD,:,2] = MASS_O2_IN_ANODE
        if PARALLEL_FLOW:
            self.m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,:,0] = MASS_H2O_IN_CATHODE
            self.m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,:,1] = MASS_H2_IN_CATHODE
        else:
            self.m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,:,0] = -MASS_H2O_IN_CATHODE
            self.m_y[INDEX_CATHODE_FLOWFIELD_BACKSIDE,:,1] = -MASS_H2_IN_CATHODE


    def setCenter(self) -> None:
        # sets the _c variables which contain only the values within the boundaries
        self.m_x_c = self.m_x[1:-1,1:-1,:].copy()
        self.m_y_c = self.m_y[1:-1,1:-1,:].copy()
        
    # def setSourceTermEnthalpie(self, S_x:float = 0,S_y:float = 0) -> None:
    #     # Sets the source term for the specific enthalpie
    #     # Args:
    #     #   S_x: source term in x direction
    #     #   S_y: source term in x direction

    #     self.S_h_x=S_x
    #     self.S_h_y=S_y

    def setSourceTermMassflow(self, S_x:float = 0,S_y:float = 0) -> None:
        # Sets the source term for the mass flow
        # Args:
        #   S_x: source term in x direction
        #   S_y: source term in x direction

        self.S_m_x=S_x
        self.S_m_y=S_y

    
    def setMassfow_x_Boundary(self, left:boundary, top:boundary, right:boundary, bottom:boundary) -> None:
        # Sets boundary condition to massflow values in x_in direction
        # Args:
        #   left: boundary object for the left boundary
        #   top: boundary object for the top boundary
        #   right: boundary object for the right boundary
        #   bottom: boundary object for the bottom boundary


        if(left.type=="D"):
            self.m_x[0,:,:]=left.value
        elif(left.type=="N"):
            self.m_x[0,:,:]=-left.value*self.space.dx[0,:]+self.m_x[1,:,:]

        if(top.type=="D"):
            self.m_x[:,-1,:]=top.value
        elif(top.type=="N"):
            self.m_x[:,-1,:]=-top.value*self.space.dy[:,-2]+self.m_x[:,-2,:]        
        
        if(right.type=="D"):
            self.m_x[-1,:,:]=right.value
        elif(right.type=="N"):
            self.m_x[-1,:,:]=right.value*self.space.dx[-2,:]+self.m_x[-2,:,:]
            
        if(bottom.type=="D"):
            self.m_x[:,0,:]=bottom.value
        elif(bottom.type=="N"):
            self.m_x[:,0,:]=bottom.value*self.space.dy[:,0]+self.m_x[:,1,:]

    def setMassfow_y_Boundary(self, left:boundary, top:boundary, right:boundary, bottom:boundary) -> None:
        # Sets boundary condition to massflow values in y direction
        # Args:
        #   left: boundary object for the left boundary
        #   top: boundary object for the top boundary
        #   right: boundary object for the right boundary
        #   bottom: boundary object for the bottom boundary


        if(left.type=="D"):
            self.m_y[0,:,:]=left.value
        elif(left.type=="N"):
            self.m_y[0,:,:]=-left.value*self.space.dx[0,:]+self.m_y[1,:,:]

        if(top.type=="D"):
            self.m_y[:,-1,:]=top.value
        elif(top.type=="N"):
            self.m_y[:,-1,:]=-top.value*self.space.dy[:,-2]+self.m_y[:,-2,:]        
        
        if(right.type=="D"):
            self.m_y[-1,:,:]=right.value
        elif(right.type=="N"):
            self.m_y[-1,:,:]=right.value*self.space.dx[-2,:]+self.m_y[-2,:,:]
            
        if(bottom.type=="D"):
            self.m_y[:,0,:]=bottom.value
        elif(bottom.type=="N"):
            self.m_y[:,0,:]=bottom.value*self.space.dy[:,0]+self.m_y[:,1,:]




class temperatureSpace: 
    # takes care of the temperature distribution

    def __init__(self, space:space) -> None:
        self.space = space

    # hanldes the variables in the mesh regarding the flow
    def createMesh(self, rowpts, colpts:int) -> None: 
        # Creates the mesh
        # Args:
        #   rowpts: number of cells in row direction
        #   colpts: number of cells in col direction
        self.rowpts = rowpts
        self.colpts = colpts

        # create temperature matrix 
        self.t = np.ones((self.rowpts+2, self.colpts+2)) 
        self.t_c = np.ones((self.rowpts,self.colpts))
        self.t_c_last = np.ones((self.rowpts,self.colpts))

        # create thermal conductivity matrix
        self.k_x = np.ones((self.rowpts+2,self.colpts+2))
        self.k_x[1:ROWPTS+1, :] = np.full((COLPTS+2,ROWPTS),THERMAL_CONDUCTIVITY_X).transpose()
        self.k_x[0,:] = THERMAL_CONDUCTIVITY_X[-1] # set to value of virtual cells
        self.k_x[-1,:] = THERMAL_CONDUCTIVITY_X[-1] # conductivity not used but must be smaller than minimal cell size not to influence dt

        self.k_y = np.ones((self.rowpts+2,self.colpts+2))
        self.k_y[1:ROWPTS+1, 1:COLPTS+1] = np.full((COLPTS,ROWPTS),THERMAL_CONDUCTIVITY_Y).transpose()
        self.k_y[0,:] = THERMAL_CONDUCTIVITY_Y[-1] # set to value of virtual cells
        self.k_y[-1,:] = THERMAL_CONDUCTIVITY_Y[-1] # conductivity not used but must be smaller than minimal cell size not to influence dt
        self.k_y[:,0] = THERMAL_CONDUCTIVITY_Y[-1] # set to value of virtual cells
        self.k_y[:,-1] = THERMAL_CONDUCTIVITY_Y[-1] # conductivity not used but must be smaller than minimal cell size not to influence dt


        # initialize source term
        self.setSources()

    def initTemperature(self, T:float) -> None:
        # Initializes the pressure in all cells 
        # Args:
        #   T: Initial value to be set for the temperature 
        self.t = T*self.t
        self.t[INDEX_ANODE_FLOWFIELD,:] = T_GAS_IN_A
        self.t[INDEX_CATHODE_FLOWFIELD_BACKSIDE,:] = T_GAS_IN_C
    def setCenter(self) -> None:
        # sets the _c variables which contain only the values within the boundaries
        self.t_c_last = self.t_c.copy()
        self.t_c = self.t[1:-1,1:-1].copy()

    def setSources(self, S:list[source]=list()) -> None:
        # Sets the source terms
        # Args:
        #   S: list with sources as source objects
        self.s = np.zeros((self.rowpts+2,self.colpts+2))
        for source in S:
            self.s[source.x+1, source.y+1] = source.value

    
    def setTBoundary(self, left:boundary, top:boundary, right:boundary, bottom:boundary) -> None:
        # Sets boundary condition to temperature values
        # See: https://math.stackexchange.com/questions/2706701/neumann-boundary-conditions-in-finite-difference for derivation of Neumann boundary condition
        # Args:
        #   left: boundary object for the left boundary
        #   top: boundary object for the top boundary
        #   right: boundary object for the right boundary
        #   bottom: boundary object for the bottom boundary

        if(left.type=="D"):
            self.t[0,:]=left.value
        elif(left.type=="N"):
            self.t[0,:]=-left.value*self.space.dx[0,:]+self.t[1,:]

        if(top.type=="D"):
            self.t[:,-1]=top.value
        elif(top.type=="N"):
            self.t[:,-1]=-top.value*self.space.dy[:,-2]+self.t[:,-2]        
        
        if(right.type=="D"):
            self.t[-1,:]=right.value
        elif(right.type=="N"):
            self.t[-1,:]=right.value*self.space.dx[-2,:]+self.t[-2,:]
            
        if(bottom.type=="D"):
            self.t[:,0]=bottom.value
        elif(bottom.type=="N"):
            self.t[:,0]=bottom.value*self.space.dy[:,0]+self.t[:,1]
